"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const dependency_injection_1 = require("@typemon/dependency-injection");
const check_1 = require("@typemon/check");
const boom_1 = __importDefault(require("@hapi/boom"));
//
const metadata_key_1 = require("./metadata-key");
const parameters_1 = require("./parameters");
const middleware_1 = require("./middleware");
//
//
//
class Runner {
    constructor(target, propertyKey, globalOptions) {
        this.target = target;
        this.propertyKey = propertyKey;
        this.globalOptions = globalOptions;
        this.targetHandle = dependency_injection_1.TargetHandle.create(this.target);
        this.controllerMetadata = this.targetHandle.getMetadata(metadata_key_1.MetadataKey.ControllerMetadata);
        this.handlerMetadata = this.targetHandle.getMetadata(metadata_key_1.MetadataKey.HandlerMetadatas).get(this.propertyKey);
        this.useAPIGateway = [
            this.handlerMetadata.options.useAPIGateway,
            this.controllerMetadata.options.useAPIGateway,
            this.globalOptions.useAPIGateway
        ].every((useAPIGateway) => check_1.Check.isUndefined(useAPIGateway) || check_1.Check.isTrue(useAPIGateway));
        this.done = false;
    }
    createRootContainer(event, context) {
        const container = dependency_injection_1.Container.create();
        container.bindAll(this.globalOptions.providers || []);
        container.bindAll(this.controllerMetadata.options.providers || []);
        container.bindAll([
            {
                identifier: parameters_1.ParameterToken.EVENT,
                useValue: event,
                lifetime: dependency_injection_1.Lifetime.Singleton
            },
            {
                identifier: parameters_1.ParameterToken.CONTEXT,
                useValue: context,
                lifetime: dependency_injection_1.Lifetime.Singleton
            }
        ]);
        if (check_1.Check.isFalse(this.useAPIGateway)) {
            return container;
        }
        container.bindAll([
            parameters_1.HttpRequest.PROVIDER,
            parameters_1.HttpResponse.PROVIDER,
            parameters_1.HttpSharedContext.PROVIDER
        ]);
        return container;
    }
    createHandlerContainer(rootContainer) {
        const container = dependency_injection_1.Container.create({
            parent: rootContainer
        });
        container.bindAll(this.handlerMetadata.options.providers || []);
        return container;
    }
    executeMiddleware(middleware, parentContainer) {
        return __awaiter(this, void 0, void 0, function* () {
            const [target, providers] = middleware_1.Middleware.WithProviders.isWithProviders(middleware)
                ? [middleware.target, middleware.providers]
                : [middleware, []];
            const targetHandle = dependency_injection_1.TargetHandle.create(target);
            const metadata = targetHandle.getMetadata(metadata_key_1.MetadataKey.MiddlewareMetadata);
            const container = dependency_injection_1.Container.create({
                parent: parentContainer
            });
            let next = false;
            container.bindAll(metadata.options.providers || []);
            container.bindAll(providers);
            container.bind({
                identifier: parameters_1.ParameterToken.NEXT,
                useValue: () => {
                    next = true;
                },
                lifetime: dependency_injection_1.Lifetime.Singleton
            });
            const instance = yield container.resolve(target);
            const instanceHandle = dependency_injection_1.InstanceHandle.create(instance);
            const parameterMetadatas = instanceHandle.getParameterMetadatas('handler');
            const parameters = yield parameters_1.ParameterResolver.resolveAll(parameterMetadatas, container);
            yield instanceHandle.callProperty('handler', parameters);
            if (check_1.Check.isFalse(next)) {
                this.done = true;
            }
        });
    }
    executeMiddlewares(middlewares, parentContainer) {
        return __awaiter(this, void 0, void 0, function* () {
            for (const middleware of middlewares) {
                yield this.executeMiddleware(middleware, parentContainer);
                if (check_1.Check.isTrue(this.done)) {
                    return;
                }
            }
        });
    }
    executeHandler(container) {
        return __awaiter(this, void 0, void 0, function* () {
            const instance = yield container.resolve(this.target);
            const instanceHandle = dependency_injection_1.InstanceHandle.create(instance);
            const parameterMetadatas = instanceHandle.getParameterMetadatas(this.propertyKey);
            const parameters = yield parameters_1.ParameterResolver.resolveAll(parameterMetadatas, container);
            const result = yield instanceHandle.callProperty(this.propertyKey, parameters);
            return result;
        });
    }
    run(event, context) {
        return __awaiter(this, void 0, void 0, function* () {
            const rootContainer = this.createRootContainer(event, context);
            const handlerContainer = this.createHandlerContainer(rootContainer);
            const tasks = [];
            if (this.useAPIGateway) {
                tasks.push(() => this.executeMiddlewares(this.globalOptions.middlewares || [], rootContainer), () => this.executeMiddlewares(this.controllerMetadata.options.middlewares || [], rootContainer), () => this.executeMiddlewares(this.handlerMetadata.options.middlewares || [], handlerContainer));
            }
            tasks.push(() => this.executeHandler(handlerContainer));
            try {
                const result = yield tasks.reduce((promise, nextTask) => __awaiter(this, void 0, void 0, function* () {
                    return promise.then((result) => {
                        if (check_1.Check.isTrue(this.done)) {
                            return result;
                        }
                        return nextTask();
                    });
                }), Promise.resolve());
                if (check_1.Check.isFalse(this.useAPIGateway)) {
                    return result;
                }
                const response = yield rootContainer.get(parameters_1.HttpResponse);
                response.statusCode = this.handlerMetadata.options.statusCode || response.statusCode;
                for (const [name, value] of Object.entries(this.handlerMetadata.options.headers || {})) {
                    response.headers.append(name, value);
                }
                if (check_1.Check.isNotUndefined(result)) {
                    response.body = result;
                }
                const serializedResponse = parameters_1.HttpResponse.serialize(response);
                return serializedResponse;
            }
            catch (error) {
                if (check_1.Check.isFalse(this.useAPIGateway)) {
                    return error;
                }
                const boom = boom_1.default.isBoom(error)
                    ? error
                    : boom_1.default.internal();
                const serializedResponse = parameters_1.HttpResponse.serialize({
                    statusCode: boom.output.statusCode,
                    headers: boom.output.headers,
                    body: {
                        statusCode: boom.output.payload.statusCode,
                        error: boom.output.payload.error,
                        message: boom.output.payload.message,
                        data: boom.data
                    }
                });
                return serializedResponse;
            }
        });
    }
}
exports.Runner = Runner;
